/***************************************************************************
                          molecule.h  -  description
                             -------------------
    begin                : Sat Jan 22 2000
    copyright            : (C) 2000 - 2003 by M. Hoverath
    email                : mario.hoverath@uni-koeln.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef _CMOLECULE
#define _CMOLECULE

#define LB_NOLABEL 0
#define LB_ATOMSYMBOL 1
#define LB_ORDINALNUMBER 2
#define LB_USER 3

#define X_AXIS 1
#define Y_AXIS 2
#define Z_AXIS 3
#define VIEW_RIGHT 1
#define VIEW_TOP 2
#define VIEW_FRONT 3
#define X_COMPONENT 1
#define Y_COMPONENT 2
#define Z_COMPONENT 3

#define DEFAULT_BONDFUDGE 1.12
#define DEFAULT_BONDCOLOR 0x202020 /* = 2105376 = RGB: 32, 32, 32 */
#define DEFAULT_PERSPECTIVEFACTOR 2.5
#define MAX_PERSPECTIVEFACTOR 10.0
#define DEFAULT_ZOOMFACTOR 1.0
#define DEFAULT_RADIUSFACTOR 0.5
#define DEFAULT_LABELTYPE LB_NOLABEL
#define DEFAULT_VIEW VIEW_FRONT
#define DEFAULT_BGCOLOR 0xFFFFFF  /* = 16777215 = RGB: 255, 255, 255 */

// default is 10 Ang.
#define MOL_MEASUREMENT 10

/* include files for QT */
#include <qwidget.h>
#include <qstring.h>
#include <qcolor.h>
#include <qpainter.h>
#include <qpixmap.h>
#include <qrect.h>

/* application specific includes */
#include <cmath>
#include <cstdlib>

/* constants */
#define PI 3.141592654

typedef struct
{
  float x;
  float y;
  float z;
  int OrdinalNumber;
  int Label;
  float Radius;
  QColor Color;
  int Pattern;
} ELEMENT;

typedef struct
{
  float x;
  float y;
  float z;
} POINT3D;

typedef struct
{
  float x;
  float y;
} POINT2D;

typedef struct
{
  int x;
  int y;
} INTPOINT2D;

/* class CPeriodicSystem */
class CPeriodicSystem : public QObject
{
  Q_OBJECT

public:
  /* constructor */
  CPeriodicSystem(QObject *parent=0, const char *name=0);

  /* destructor */
  ~CPeriodicSystem();

  /* operators */
  CPeriodicSystem& operator=(CPeriodicSystem &pse);

  /* methods */
  void RestoreData(void); 

  /* get ... */
  QString GetSymbol(int ordinalNumber);
  int GetOrdinalNumber(QString Symbol);
  float GetRadius(int ordinalNumber);
  QColor GetColor(int ordinalNumber);
  int GetPattern(int ordinalNumber);

  /* set ... */
  int SetRadius(int ordinalNumber, float radius);
  int SetColor(int ordinalNumber, QColor color);
  int SetPattern(int ordinalNumber, int pattern);

protected: 
  QString Symbol[109];
  float Radius[109]; 
  QColor Color[109];
  int Pattern[109];
  QColor tmpColor;
};

/* class CCoordinateSystem */
class CCoordinateSystem : public QObject
{
  Q_OBJECT

public:
  /* constructor */
  CCoordinateSystem(QObject *parent=0, const char *name=0);

  /* destructor */
  ~CCoordinateSystem();

  /* operators */
  CCoordinateSystem& operator=(CCoordinateSystem &cs);

  /* methods */
  int Rotate(int axis, float angle);
  int CorrectRoundingErrors(void);
  void Reset(void);
  
  /* get ... */
  POINT3D GetVector(int axis);
  POINT3D GetInverseVector(int axis);
  float GetMatrix(int axis, int component);

  /* set ... */
  int SetMatrix(float ix, float iy, float iz,
                float jx, float jy, float jz,
                float kx, float ky, float kz,
                int correctErrors = 1);

protected:
  /* members */
  POINT3D I;
  POINT3D J;
  POINT3D K;
  POINT3D tmp_point3d;
};

/* class CMolecule */
class CMolecule : public QObject
{
  Q_OBJECT

public: 
  /* constructor */
  CMolecule(QObject *parent=0, const char *name=0);

  /* destructor */
  ~CMolecule(); 
 
  /* operators */
  CMolecule& operator=(CMolecule &Mol);

  /* methods */
  void Serialize();
  int Rotate(int axis, float angle);
  void Translate(float transX, float transY, float transZ);
  void Translate(POINT3D trans);
  void TranslateAbsolute(int DeltaPixelPosX, int DeltaPixelPosY, int View = VIEW_FRONT);
  void ResetMovements(void);
  void ResetCoordinateSystem(void);
  void ResetTranslation(void);
  int InitMolecule(void);
  int Capture(int MousePosX, int MousePosY, int View = VIEW_FRONT);
  int DrawStructure(QPaintDevice *pdev, const QRect &rectClient);

  /* add, delete ... */
  int AddAtom(ELEMENT El, int MakeBonds = 1);
  int DeleteAtom(int Atom);
  int DeleteAtomByLabel(int LabelAtom);
  int AddBond(int Atom1, int Atom2);
  int AddBondByLabels(int LabelAtom1, int LabelAtom2);
  int DeleteBond(int Atom1, int Atom2);
  int DeleteBondByLabels(int LabelAtom1, int LabelAtom2);
  int DeleteMolecule();

  /* get ... */
  POINT3D GetCoordinateSystem(int axis);
  float GetMatrix(int axis, int component);
  POINT3D GetTranslation(void);
  POINT3D GetAtomCoordinate(int Atom, int ConsiderRotation = 0, int ConsiderTranslation = 0);
  INTPOINT2D GetBond(int Bond);
  int GetOrdinalNumber(int Atom);
  int GetLabel(int Atom);
  int GetAtomNumberByLabel(int LabelAtom);
  float GetRadius(int Atom);
  QColor GetColor(int Atom);
  int GetPattern(int Atom);
  QString GetTitle(void);
  int GetNumberOfAtoms(void);
  int GetNumberOfBonds(void);
  float GetBondFudge(void);
  QColor GetBondColor(void);
  float GetRadiusFactor(void);
  float GetPerspectiveFactor(void);
  float GetZoomFactor(void);
  int GetLabelType(void);
  int GetView(void);
  QColor GetBackgroundColor(void);

  /* set ... */
  int SetMatrix(float ix, float iy, float iz,
                float jx, float jy, float jz,
                float kx, float ky, float kz,
                int correctErrors = 1);
  int SetAtomCoordinate(int Atom, POINT3D NewCoordinate, int updateBonds = 0);
  int SetTranslation(float transX, float transY, float transZ);
  int SetAbsoluteAtomCoordinate(int Atom, int PixelPosX, int PixelPosY, int updateBonds = 0, int View = VIEW_FRONT);
  int SetOrdinalNumber(int Atom, int NewOrdinalNumber);
  int SetLabel(int Atom, int NewLabel);
  int SetRadius(int Atom, float NewRadius);
  int SetRadius(float NewRadius);
  int SetColor(int Atom, QColor NewColor);
  int SetColor(QColor NewColor);
  int SetPattern(int Atom, int NewPattern);
  int SetPattern(int NewPattern);
  int SetTitle(QString NewTitle);
  int SetBondFudge(float NewBondFudge, int updateBonds = 1);
  int SetBondColor(QColor NewBondColor);
  int SetRadiusFactor(float NewRadiusFactor);
  int SetPerspectiveFactor(float NewPerspectiveFactor);
  int SetZoomFactor(float NewZoomFactor);
  int SetLabelType(int NewLabelType);
  int SetView(int NewView = VIEW_FRONT);
  int SetBackgroundColor(QColor newBGColor);

protected:
  int execRotationTranslation(ELEMENT *in, ELEMENT *out, int view = 0);
  int execRotation(POINT3D &in, POINT3D &out, int view = 0);
  int execInvRotation(POINT3D &in, POINT3D &out, int view = 0);
  float calcPerspective(float &PosZ);
  int execPerspective(ELEMENT *elements);
  int execInvPerspective(float &posX, float &posY, float &posZ);
  int sortAtoms(ELEMENT *elements, INTPOINT2D *bonds);
  POINT2D mapFromWidgetPoint(int posX, int posY);

protected:
  /* dynamic fields */
  ELEMENT *Element, *help_Element;
  int SizeOfElementField;
  INTPOINT2D *BondPartners, *help_BondPartners;
  int SizeOfBondPartnersField;

  /* members */
  CCoordinateSystem cs;
  CPeriodicSystem pse;
  QString Title; 
  int NumberOfAtoms;
  int NumberOfBonds;
  float BondFudge; 
  QColor BondColor;
  QColor bgColor;
  QRect usedClientRect;
  float PerspectiveFactor;
  float MaxAtomCenterDistance;
  float ZoomFactor;
  float RadiusFactor;
  POINT3D translation;
  int LabelType;
  int View;
};

#endif /* _CMOLECULE */
